(***********************************************************************)
(*                                                                     *)
(*                       Ambients in JoCaml                            *)
(*                                                                     *)
(*               Alan Schmitt, projet Para, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1999 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

{
open Parser;; (* type Token defined in Parser.mli *)
exception Eof;;
exception Illegal;;

type error =
    Illegal_character
  | Unterminated_comment
  | Unterminated_string

exception Error of error * int * int

(* To buffer string literals *)

let initial_string_buffer = String.create 256
let string_buff = ref initial_string_buffer
let string_index = ref 0

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0

let store_string_char c =
  if !string_index >= String.length (!string_buff) then begin
    let new_buff = String.create (String.length (!string_buff) * 2) in
      String.blit (!string_buff) 0 new_buff 0 (String.length (!string_buff));
      string_buff := new_buff
  end;
  String.unsafe_set (!string_buff) (!string_index) c;
  incr string_index

let get_stored_string () =
  let s = String.sub (!string_buff) 0 (!string_index) in
  string_buff := initial_string_buffer;
  s

(* To translate escape sequences *)

let char_for_backslash =
  match Sys.os_type with
  | "Unix" | "Win32" ->
      begin function
      | 'n' -> '\010'
      | 'r' -> '\013'
      | 'b' -> '\008'
      | 't' -> '\009'
      | c   -> c
      end
  | "MacOS" ->
      begin function
      | 'n' -> '\013'
      | 'r' -> '\010'
      | 'b' -> '\008'
      | 't' -> '\009'
      | c   -> c
      end
  | x -> print_string "Lexer: unknown system type"; exit 1

let char_for_decimal_code lexbuf i =
  let c = 100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
           10 * (Char.code(Lexing.lexeme_char lexbuf (i+1)) - 48) +
                (Char.code(Lexing.lexeme_char lexbuf (i+2)) - 48) in  
  Char.chr(c land 0xFF)

(* To store the position of the beginning of a string and comment *)
let string_start_pos = ref 0

}
rule token = parse
  [' ' '\t' '\n']
                  { token lexbuf }     (* skip blanks *)
| "|"             { PAR }
| "in"            { IN }
| "out"           { OUT }
| "open"          { OPEN }
| "("             { LPAREN }
| ")"             { RPAREN }
| "["             { LBRACKET }
| "]"             { RBRACKET }
| "<"             { LPOINT }
| ">"             { RPOINT } 
| "."             { DOT }
| "nu"            { NEW }
| "!"             { BANG }
| "0"             { ZERO }
| ";;"            { SEMISEMI }
| "$"             { DOLLAR }
| ['A'-'Z''a'-'z'] + ['0'-'9'] *
                   { IDENT (Lexing.lexeme lexbuf) }
| "\""
    { reset_string_buffer();
      let string_start = Lexing.lexeme_start lexbuf in
      string_start_pos := string_start;
      string lexbuf;
      lexbuf.Lexing.lex_start_pos <-
        string_start - lexbuf.Lexing.lex_abs_pos;
      STRING (get_stored_string()) }
| eof             { raise Eof }
| _               { raise Illegal }    

and string = parse
    '"'
      { () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *
      { string lexbuf }
  | '\\' ['\\' '"' 'n' 't' 'b' 'r']
      { store_string_char(char_for_backslash(Lexing.lexeme_char lexbuf 1));
        string lexbuf }
  | '\\' ['0'-'9'] ['0'-'9'] ['0'-'9']
      { store_string_char(char_for_decimal_code lexbuf 1);
         string lexbuf }
  | eof
      { raise (Error (Unterminated_string,
                      !string_start_pos, !string_start_pos+1)) }
  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        string lexbuf }
